/*
   Audio.cpp

    This library plays ONLY mp3 files from SD card

    Created on: Oct 26,2018
    Updated on: Oct 15,2020
        Author: Wolle

    Modifications and Implementation: Craig A. Lindley
    Last Update: 12/19/2024
*/

#include "BluetoothA2DPSource.h"

#include "AudioPlayer.h"
#include "mp3_decoder.h"

//---------------------------------------------------------------------------------------------------------------------
AudioBuffer::AudioBuffer() {
}

AudioBuffer::~AudioBuffer() {
  if (m_buffer) free(m_buffer);
  m_buffer = NULL;
}

size_t AudioBuffer::init() {
  if (psramInit()) {
    // PSRAM found, AudioBuffer will be allocated in PSRAM
    m_buffSize = m_buffSizePSRAM;
    if (m_buffer == NULL) {
      m_buffer = (uint8_t*)ps_calloc(m_buffSize, sizeof(uint8_t));
      m_buffSize = m_buffSizePSRAM - m_resBuffSize;
      if (m_buffer == NULL) {
        // not enough space in PSRAM, use ESP32 Flash Memory instead
        m_buffer = (uint8_t*)calloc(m_buffSize, sizeof(uint8_t));
        m_buffSize = m_buffSizeRAM - m_resBuffSize;
      }
    }
  }
  else { // no PSRAM available, use ESP32 Flash Memory"
    m_buffSize = m_buffSizeRAM;
    m_buffer = (uint8_t*)calloc(m_buffSize, sizeof(uint8_t));
    m_buffSize = m_buffSizeRAM - m_resBuffSize;
  }
  if (!m_buffer) return 0;
  resetBuffer();
  return m_buffSize;
}

size_t AudioBuffer::freeSpace() {
  if (m_readPtr >= m_writePtr) {
    m_freeSpace = (m_readPtr - m_writePtr);
  }
  else {
    m_freeSpace = (m_endPtr - m_writePtr) + (m_readPtr - m_buffer);
  }
  if (m_f_start) m_freeSpace = m_buffSize;
  return m_freeSpace - 1;
}

size_t AudioBuffer::writeSpace() {
  if (m_readPtr >= m_writePtr) {
    m_writeSpace = (m_readPtr - m_writePtr - 1); // readPtr must not be overtaken
  }
  else {
    if (getReadPos() == 0) m_writeSpace = (m_endPtr - m_writePtr - 1);
    else m_writeSpace = (m_endPtr - m_writePtr);
  }
  if (m_f_start) m_writeSpace = m_buffSize - 1;
  return m_writeSpace;
}

size_t AudioBuffer::bufferFilled() {
  if (m_writePtr >= m_readPtr) {
    m_dataLength = (m_writePtr - m_readPtr);
  }
  else {
    m_dataLength = (m_endPtr - m_readPtr) + (m_writePtr - m_buffer);
  }
  return m_dataLength;
}

void AudioBuffer::bytesWritten(size_t bw) {
  m_writePtr += bw;
  if (m_writePtr == m_endPtr) {
    m_writePtr = m_buffer;
  }
  if (bw && m_f_start) m_f_start = false;
}

void AudioBuffer::bytesWasRead(size_t br) {
  m_readPtr += br;
  if (m_readPtr >= m_endPtr) {
    size_t tmp = m_readPtr - m_endPtr;
    m_readPtr = m_buffer + tmp;
  }
}

uint8_t* AudioBuffer::writePtr() {
  return m_writePtr;
}

uint8_t* AudioBuffer::readPtr() {
  size_t len = m_endPtr - m_readPtr;
  if (len < 1600) { // be sure the last frame is completed
    memcpy(m_endPtr, m_buffer,  1600);
  }
  return m_readPtr;
}

void AudioBuffer::resetBuffer() {
  m_writePtr = m_buffer;
  m_readPtr  = m_buffer;
  m_endPtr   = m_buffer + m_buffSize;
  m_f_start  = true;
}

uint32_t AudioBuffer::getWritePos() {
  return m_writePtr - m_buffer;
}

uint32_t AudioBuffer::getReadPos() {
  return m_readPtr - m_buffer;
}
//---------------------------------------------------------------------------------------------------------------------
// CAL - Removed from AudioPlayer.h so C and C++ code can interface

bool    m_f_playing = false;    // valid mp3 stream recognized
int16_t m_outBuff[2048 * 2];    // [1152 * 2] interleaved L/R
int16_t m_validSamples = 0;
int16_t m_curSample = 0;
uint8_t m_vol = 17;             // Default volume
uint8_t m_volIndex = 10;

const uint8_t volumetable[22] = {
  0,  1,  2,  3,  4 , 6 , 8, 10, 12, 14, 17,
  20, 23, 27, 30 , 34, 38, 43 , 48, 52, 58, 64
}; //22 elements

//---------------------------------------------------------------------------------------------------------------------

// Increment the volume level
void volumeUp() {

  m_volIndex += 2;

  if (m_volIndex > 21) {
    m_volIndex = 21;
  }
  m_vol = volumetable[m_volIndex];
}

// Decrement the volume level
void volumeDown() {

  m_volIndex -= 2;

  if (m_volIndex < 0) {
    m_volIndex = 0;
  }
  m_vol = volumetable[m_volIndex];
}

//---------------------------------------------------------------------------------------------------------------------
int16_t Gain(int16_t s) {
  int32_t v;
  v = (s * m_vol) >> 6;
  return (int16_t)(v & 0xffff);
}

//---------------------------------------------------------------------------------------------------------------------

// CAL
bool AudioPlayer::playChunk() {
  // If we've got data, try and pump it out..
  // Assumes 16 bit stereo samples
  // Note: this method will be called when a file is
  // playing everytime m_validSamples goes to 0.
  delay(1);
  return true;
}

// CAL - Function added to feed sample data to bluetooth
// NOTE: bluetooth requests 512 bytes or 128 stereo samples
//       MP3 provides 1152 stereo samples
int32_t bluetoothDataProvider(Channels *channels, int32_t len) {

  //  Serial.printf("m_validSamples: %d\n", m_validSamples);

  // Are there valid stereo samples ?
  if (m_validSamples != 0) {
    // Yes there are valid samples
    for (int i = 0; i < len && (m_validSamples > 0); i++) {
      channels[i].channel1 = Gain(m_outBuff[m_curSample * 2]);
      channels[i].channel2 = Gain(m_outBuff[m_curSample * 2 + 1]);
      m_validSamples--;
      m_curSample++;
    }
    while (m_validSamples == 0) {
      delay(1);
    }
  } else {
    // No valid samples so output zeros for silence
    for (int i = 0; i < len; i++) {
      channels[i].channel1 = 0;
      channels[i].channel2 = 0;
    }
  }
  return len;
}

//---------------------------------------------------------------------------------------------------------------------

AudioPlayer::AudioPlayer() {

  size_t size = InBuff.init();
  if (size == m_buffSizeRAM   - m_resBuffSize) {
    sprintf(chbuf, "PSRAM not found, inputBufferSize = %u bytes", size - 1);
    if (audio_info) audio_info(chbuf);
    m_f_psram = false;
  }
  if (size == m_buffSizePSRAM - m_resBuffSize) {
    sprintf(chbuf, "PSRAM found, inputBufferSize = %u bytes", size - 1);
    if (audio_info) audio_info(chbuf);
    m_f_psram = true;
  }
}

//---------------------------------------------------------------------------------------------------------------------

AudioPlayer::~AudioPlayer() {
  InBuff.~AudioBuffer();
}

//---------------------------------------------------------------------------------------------------------------------

void AudioPlayer::reset() {

  stopSong();
  InBuff.resetBuffer();
  MP3Decoder_FreeBuffers();

  sprintf(chbuf, "buffers freed, free Heap: %u bytes", ESP.getFreeHeap());
  if (audio_info) audio_info(chbuf);

  m_f_localfile = false;                                  // SD connected
  m_f_playing = false;
  m_audioCurrentTime = 0;                                 // Reset playtimer
  m_audioFileDuration = 0;
  m_avr_bitrate = 0;                                      // the same as m_bitrate if CBR, median if VBR
  m_bitRate = 0;                                          // Bitrate still unknown
  m_codec = CODEC_NONE;
  m_curSample = 0;
}

//-----------------------------------------------------------------------------------------------------------------------------------
bool AudioPlayer::connecttoSD(String sdfile) {
  return connecttoFS(SD, sdfile);
}
//-----------------------------------------------------------------------------------------------------------------------------------
bool AudioPlayer::connecttoFS(fs::FS &fs, String file) {

  reset(); // free buffers an ser defaults

  uint16_t i = 0, s = 0;
  m_f_localfile = true;

  m_audioName = file.substring(file.lastIndexOf('/') + 1, file.length());
  sprintf(chbuf, "Reading file: %s", m_audioName.c_str());
  if (audio_info) audio_info(chbuf);
  audiofile = fs.open(file);
  if (!audiofile) {
    if (audio_info) audio_info("Failed to open file for reading");
    return false;
  }
  String afn = (String)audiofile.name();  //audioFileName
  if (afn.endsWith(".mp3") || afn.endsWith(".MP3")) { // MP3 section
    m_codec = CODEC_MP3;
    MP3Decoder_AllocateBuffers();
    sprintf(chbuf, "MP3Decoder has been initialized, free Heap: %u bytes", ESP.getFreeHeap());
    if (audio_info) audio_info(chbuf);
    audiofile.readBytes(chbuf, 10);
    if ((chbuf[0] != 'I') || (chbuf[1] != 'D') || (chbuf[2] != '3')) {
      if (audio_info) audio_info("file has no mp3 tag, skip metadata");
      setFilePos(0);
      m_f_running = true;
      return false;
    }
    m_rev = chbuf[3];
    switch (m_rev) {
      case 2:
        m_f_unsync = (chbuf[5] & 0x80);
        m_f_exthdr = false;
        break;
      case 3:
      case 4:
        m_f_unsync = (chbuf[5] & 0x80); // bit7
        m_f_exthdr = (chbuf[5] & 0x40); // bit6 extended header
        break;
    };

    m_id3Size  = chbuf[6]; m_id3Size = m_id3Size << 7;
    m_id3Size |= chbuf[7]; m_id3Size = m_id3Size << 7;
    m_id3Size |= chbuf[8]; m_id3Size = m_id3Size << 7;
    m_id3Size |= chbuf[9];

    // Every read from now may be unsync'd
    sprintf(chbuf, "ID3 version=%i", m_rev);
    if (audio_info) audio_info(chbuf);
    sprintf(chbuf, "ID3 framesSize=%i", m_id3Size);
    if (audio_info) audio_info(chbuf);
    readID3Metadata();
    m_f_running = true;
    return true;
  } // end MP3 section

  if (audio_info) audio_info("Not mp3 format");
  return false;
}

//---------------------------------------------------------------------------------------------------------------------
void AudioPlayer::readID3Metadata() {
  char frameid[5];
  int framesize = 0;
  bool compressed;
  char value[256];
  bool bitorder = false;
  uint8_t uni_h = 0;
  uint8_t uni_l = 0;
  int id3Size = m_id3Size;
  String tag = "";
  if (m_f_exthdr) {
    if (audio_info) audio_info("ID3 extended header");
    int ehsz = (audiofile.read() << 24) | (audiofile.read() << 16) | (audiofile.read() << 8) | (audiofile.read());
    id3Size -= 4;
    for (int j = 0; j < ehsz - 4; j++) {
      audiofile.read();
      id3Size--;
    } // Throw it away
  } else if (audio_info) audio_info("ID3 normal frames");

  do {
    frameid[0] = audiofile.read();
    frameid[1] = audiofile.read();
    frameid[2] = audiofile.read();
    id3Size -= 3;
    if   (m_rev == 2) frameid[3] = 0;
    else {
      frameid[3] = audiofile.read();
      id3Size--;
    }
    frameid[4] = 0; // terminate the string
    tag = frameid;
    if (frameid[0] == 0 && frameid[1] == 0 && frameid[2] == 0 && frameid[3] == 0) {
      // We're in padding
      while (id3Size != 0) {
        audiofile.read();
        id3Size--;
      }
    }
    else {
      if (m_rev == 2) {
        framesize = (audiofile.read() << 16) | (audiofile.read() << 8) | (audiofile.read());
        id3Size -= 3;
        compressed = false;
      }
      else {
        framesize = (audiofile.read() << 24) | (audiofile.read() << 16) | (audiofile.read() << 8) | (audiofile.read());
        id3Size -= 4;
        audiofile.read(); // skip 1st flag
        id3Size--;
        compressed = audiofile.read() & 0x80;
        id3Size--;
      }
      if (compressed) {
        log_i("iscompressed");
        int decompsize = (audiofile.read() << 24) | (audiofile.read() << 16) | (audiofile.read() << 8) | (audiofile.read());
        id3Size -= 4;
        (void) decompsize;
        for (int j = 0; j < framesize; j++) {
          audiofile.read();
          id3Size--;
        }
      }
      // Read the value
      uint32_t i = 0; uint16_t j = 0, k = 0, m = 0;
      bool isUnicode;
      if (framesize > 0) {
        isUnicode = (audiofile.read() == 1) ? true : false;
        id3Size--;
        if (framesize < 256) {
          audiofile.readBytes(value, framesize - 1); id3Size -= framesize - 1;
          i = framesize - 1; value[framesize - 1] = 0;
        }
        else {
          if (tag == "APIC") { // a image embedded in file, passing it to external function
            //log_i("it's a image");
            isUnicode = false;
            const uint32_t preReadFilePos = getFilePos();
            if (audio_id3image) audio_id3image(audiofile, framesize);
            setFilePos(preReadFilePos + framesize - 1);
            id3Size -= framesize - 1;
          }
          else {
            // store the first 255 bytes in buffer and cut the remains
            audiofile.readBytes(value, 255); id3Size -= 255;
            value[255] = 0; i = 255;
            // big block, skip it
            setFilePos(getFilePos() + framesize - 1 - 255); id3Size -= framesize - 1;
          }
        }
        if (isUnicode && framesize > 1) { // convert unicode to utf-8 U+0020...U+07FF
          j = 0; m = 0;
          while (m < i - 1) {
            if ((value[m] == 0xFE) && (value[m + 1] == 0xFF)) {
              bitorder = true;  // MSB/LSB
              j = m + 2;
            }
            if ((value[m] == 0xFF) && (value[m + 1] == 0xFE)) {
              bitorder = false;  //LSB/MSB
              j = m + 2;
            }
            m++;
          } // seek for last bitorder
          m = 0;
          if (j > 0) {
            for (k = j; k < i - 1; k += 2) {
              if (bitorder == true) {
                uni_h = value[k];
                uni_l = value[k + 1];
              }
              else              {
                uni_l = value[k];
                uni_h = value[k + 1];
              }
              uint16_t uni_hl = (uni_h << 8) + uni_l;
              uint8_t utf8_h = (uni_hl >> 6); // div64
              uint8_t utf8_l = uni_l;
              if (utf8_h > 3) {
                utf8_h += 0xC0;
                if     (uni_l < 0x40) utf8_l = uni_l + 0x80;
                else if (uni_l < 0x80) utf8_l = uni_l += 0x40;
                else if (uni_l < 0xC0) utf8_l = uni_l;
                else    utf8_l = uni_l - 0x40;
              }
              if (utf8_h > 3) {
                value[m] = utf8_h;
                m++;
              }
              value[m] = utf8_l; m++;
            }
          } value[m] = 0; i = m;
        }
      }
      chbuf[0] = 0; j = 0; k = 0;
      while (j < i) {
        if (value[j] > 0x19) {
          value[k] = value[j];  //remove non printables
          k++;
        } else {
          i--;
        } j++;
      }
      value[i] = 0; // new termination
      // Revision 2
      if (tag == "CNT") sprintf(chbuf, "Play counter: %s", value);
      if (tag == "COM") sprintf(chbuf, "Comments: %s", value);
      if (tag == "CRA") sprintf(chbuf, "Audio encryption: %s", value);
      if (tag == "CRM") sprintf(chbuf, "Encrypted meta frame: %s", value);
      if (tag == "ETC") sprintf(chbuf, "Event timing codes: %s", value);
      if (tag == "EQU") sprintf(chbuf, "Equalization: %s", value);
      if (tag == "IPL") sprintf(chbuf, "Involved people list: %s", value);
      if (tag == "PIC") sprintf(chbuf, "Attached picture: %s", value);
      if (tag == "SLT") sprintf(chbuf, "Synchronized lyric/text: %s", value);
      if (tag == "TAL") sprintf(chbuf, "Album/Movie/Show title: %s", value);
      if (tag == "TBP") sprintf(chbuf, "BPM (Beats Per Minute): %s", value);
      if (tag == "TCM") sprintf(chbuf, "Composer: %s", value);
      if (tag == "TCO") sprintf(chbuf, "Content type: %s", value);
      if (tag == "TCR") sprintf(chbuf, "Copyright message: %s", value);
      if (tag == "TDA") sprintf(chbuf, "Date: %s", value);
      if (tag == "TDY") sprintf(chbuf, "Playlist delay: %s", value);
      if (tag == "TEN") sprintf(chbuf, "Encoded by: %s", value);
      if (tag == "TFT") sprintf(chbuf, "File type: %s", value);
      if (tag == "TIM") sprintf(chbuf, "Time: %s", value);
      if (tag == "TKE") sprintf(chbuf, "Initial key: %s", value);
      if (tag == "TLA") sprintf(chbuf, "Language(s): %s", value);
      if (tag == "TLE") sprintf(chbuf, "Length: %s", value);
      if (tag == "TMT") sprintf(chbuf, "Media type: %s", value);
      if (tag == "TOA") sprintf(chbuf, "Original artist(s)/performer(s): %s", value);
      if (tag == "TOF") sprintf(chbuf, "Original filename: %s", value);
      if (tag == "TOL") sprintf(chbuf, "Original Lyricist(s)/text writer(s): %s", value);
      if (tag == "TOR") sprintf(chbuf, "Original release year: %s", value);
      if (tag == "TOT") sprintf(chbuf, "Original album/Movie/Show title: %s", value);
      if (tag == "TP1") sprintf(chbuf, "Lead artist(s)/Lead performer(s)/Soloist(s)/Performing group: %s", value);
      if (tag == "TP2") sprintf(chbuf, "Band/Orchestra/Accompaniment: %s", value);
      if (tag == "TP3") sprintf(chbuf, "Conductor/Performer refinement: %s", value);
      if (tag == "TP4") sprintf(chbuf, "Interpreted, remixed, or otherwise modified by: %s", value);
      if (tag == "TPA") sprintf(chbuf, "Part of a set: %s", value);
      if (tag == "TPB") sprintf(chbuf, "Publisher: %s", value);
      if (tag == "TRC") sprintf(chbuf, "ISRC (International Standard Recording Code): %s", value);
      if (tag == "TRD") sprintf(chbuf, "Recording dates: %s", value);
      if (tag == "TRK") sprintf(chbuf, "Track number/Position in set: %s", value);
      if (tag == "TSI") sprintf(chbuf, "Size: %s", value);
      if (tag == "TSS") sprintf(chbuf, "Software/hardware and settings used for encoding: %s", value);
      if (tag == "TT1") sprintf(chbuf, "Content group description: %s", value);
      if (tag == "TT2") sprintf(chbuf, "Title/Songname/Content description: %s", value);
      if (tag == "TT3") sprintf(chbuf, "Subtitle/Description refinement: %s", value);
      if (tag == "TXT") sprintf(chbuf, "Lyricist/text writer: %s", value);
      if (tag == "TXX") sprintf(chbuf, "User defined text information frame: %s", value);
      if (tag == "TYE") sprintf(chbuf, "Year: %s", value);
      if (tag == "UFI") sprintf(chbuf, "Unique file identifier: %s", value);
      if (tag == "ULT") sprintf(chbuf, "Unsychronized lyric/text transcription: %s", value);
      if (tag == "WAF") sprintf(chbuf, "Official audio file webpage: %s", value);
      if (tag == "WAR") sprintf(chbuf, "Official artist/performer webpage: %s", value);
      if (tag == "WAS") sprintf(chbuf, "Official audio source webpage: %s", value);
      if (tag == "WCM") sprintf(chbuf, "Commercial information: %s", value);
      if (tag == "WCP") sprintf(chbuf, "Copyright/Legal information: %s", value);
      if (tag == "WPB") sprintf(chbuf, "Publishers official webpage: %s", value);
      if (tag == "WXX") sprintf(chbuf, "User defined URL link frame: %s", value);
      // Revision 3
      if (tag == "COMM") sprintf(chbuf, "Comment: %s", value);
      if (tag == "OWNE") sprintf(chbuf, "Ownership: %s", value);
      if (tag == "PRIV") sprintf(chbuf, "Private: %s", value);
      if (tag == "SYLT") sprintf(chbuf, "SynLyrics: %s", value);
      if (tag == "TALB") sprintf(chbuf, "Album: %s", value);
      if (tag == "TBPM") sprintf(chbuf, "BeatsPerMinute: %s", value);
      if (tag == "TCMP") sprintf(chbuf, "Compilation: %s", value);
      if (tag == "TCOM") sprintf(chbuf, "Composer: %s", value);
      if (tag == "TCOP") sprintf(chbuf, "Copyright: %s", value);
      if (tag == "TDAT") sprintf(chbuf, "Date: %s", value);
      if (tag == "TEXT") sprintf(chbuf, "Lyricist: %s", value);
      if (tag == "TIME") sprintf(chbuf, "Time: %s", value);
      if (tag == "TIT1") sprintf(chbuf, "Grouping: %s", value);
      if (tag == "TIT2") sprintf(chbuf, "Title: %s", value);
      if (tag == "TIT3") sprintf(chbuf, "Subtitle: %s", value);
      if (tag == "TLAN") sprintf(chbuf, "Language: %s", value);
      if (tag == "TLEN") sprintf(chbuf, "Length: %s", value);
      if (tag == "TMED") sprintf(chbuf, "Media: %s", value);
      if (tag == "TOAL") sprintf(chbuf, "OriginalAlbum: %s", value);
      if (tag == "TOPE") sprintf(chbuf, "OriginalArtist: %s", value);
      if (tag == "TORY") sprintf(chbuf, "OriginalReleaseYear: %s", value);
      if (tag == "TPE1") sprintf(chbuf, "Artist: %s", value);
      if (tag == "TPE2") sprintf(chbuf, "Band: %s", value);
      if (tag == "TPE3") sprintf(chbuf, "Conductor: %s", value);
      if (tag == "TPE4") sprintf(chbuf, "InterpretedBy: %s", value);
      if (tag == "TPOS") sprintf(chbuf, "PartOfSet: %s", value);
      if (tag == "TPUB") sprintf(chbuf, "Publisher: %s", value);
      if (tag == "TRCK") sprintf(chbuf, "Track: %s", value);
      if (tag == "TRDA") sprintf(chbuf, "RecordingDates: %s", value);
      if (tag == "TXXX") sprintf(chbuf, "UserDefinedText: %s", value);
      if (tag == "TYER") sprintf(chbuf, "Year: %s", value);
      if (tag == "USER") sprintf(chbuf, "TermsOfUse: %s", value);
      if (tag == "USLT") sprintf(chbuf, "Lyrics: %s", value);
      if (tag == "XDOR") sprintf(chbuf, "OriginalReleaseTime: %s", value);
      if (chbuf[0] != 0) if (audio_id3data) audio_id3data(chbuf);
    }
  } while (id3Size > 0);
}
//---------------------------------------------------------------------------------------------------------------------
void AudioPlayer::stopSong() {
  if (m_f_running) {
    m_f_running = false;
    audiofile.close();
  }
  memset(m_outBuff, 0, sizeof(m_outBuff));     //Clear OutputBuffer
}

//---------------------------------------------------------------------------------------------------------------------
bool AudioPlayer::pauseResume()
{
  bool retVal = false;
  if (m_f_localfile) {
    m_f_running = ! m_f_running;
    retVal = true;
    if (!m_f_running) {
      memset(m_outBuff, 0, sizeof(m_outBuff));               //Clear OutputBuffer
    }
  }
  return retVal;
}


//---------------------------------------------------------------------------------------------------------------------
void AudioPlayer::loop()
{
  // Play localfile
  if (m_f_localfile) { // Playing file fron SPIFFS or SD?
    processLocalFile();
  }
}
//---------------------------------------------------------------------------------------------------------------------
void AudioPlayer::processLocalFile()
{
  static boolean lastChunk = false;
  int bytesDecoded = 0;
  if (audiofile && m_f_running && m_f_localfile) {
    uint32_t bytesCanBeWritten = 0;
    uint32_t bytesCanBeRead = 0;
    int32_t  bytesAddedToBuffer = 0;

    bytesCanBeWritten = InBuff.writeSpace();
    bytesAddedToBuffer = audiofile.read(InBuff.writePtr(), bytesCanBeWritten);

    if (bytesAddedToBuffer > 0) InBuff.bytesWritten(bytesAddedToBuffer);
    bytesCanBeRead = InBuff.bufferFilled();
    if (bytesCanBeRead > 1600) bytesCanBeRead = 1600;
    if (bytesCanBeRead == 1600) { // mp3 or aac frame complete?
      if (!m_f_stream) {
        m_f_stream = true;
        if (audio_info) audio_info("stream ready");
      }
      bytesDecoded = sendBytes(InBuff.readPtr(), bytesCanBeRead);
      if (bytesDecoded > 0) InBuff.bytesWasRead(bytesDecoded);
      lastChunk = false;
      return;
    }
    if (!bytesAddedToBuffer) { // eof
      if (lastChunk == false) {
        if (bytesCanBeRead) {
          bytesDecoded = sendBytes(InBuff.readPtr(), bytesCanBeRead); // play last chunk(s)
          if (bytesDecoded > 0) InBuff.bytesWasRead(bytesDecoded);
          if (bytesDecoded < 100) { // unlikely framesize
            lastChunk = true;
            return;
          }
          return;
        }
        lastChunk = true;
        return; // release the thread, continue on the next pass
      }

      stopSong();
      m_f_stream = false;
      m_f_localfile = false;
      MP3Decoder_FreeBuffers();
      sprintf(chbuf, "End of file %s", m_audioName.c_str());
      if (audio_info) audio_info(chbuf);
      if (audio_eof_mp3) audio_eof_mp3(m_audioName.c_str());
    }
  }
}

//---------------------------------------------------------------------------------------------------------------------

int AudioPlayer::sendBytes(uint8_t *data, size_t len) {

  static uint32_t lastRet = 0, count = 0, swnf = 0;
  static uint32_t lastSampleRate = 0, lastChannels = 0, lastBitsPerSeconds = 0, lastBitRate = 0;
  int nextSync = 0;
  if (!m_f_playing) {
    lastSampleRate = 0;
    lastChannels = 0;
    lastBitsPerSeconds = 0;
    lastBitRate = 0;
    if (m_codec == CODEC_MP3) nextSync = MP3FindSyncWord(data, len);

    if (nextSync == -1) {
      swnf++; // syncword not found counter, can be multimediadata
      if (audio_info && swnf < 1) audio_info("syncword not found");
      return -1;
    }
    else {
      if (audio_info && swnf > 0) {
        sprintf(chbuf, "syncword not found %i times", swnf);
        audio_info(chbuf);
      }
      swnf = 0;
    }
    if (nextSync > 0) {
      sprintf(chbuf, "syncword found at pos %i", nextSync);
      if (audio_info) audio_info(chbuf);
      return nextSync;
    }
    if (audio_info) audio_info("syncword found at pos 0");
    count = 0;
    //        m_bitRate=0;
    m_f_playing = true;
    return nextSync;
  }

  m_bytesLeft = len;
  int ret = 0;
  int bytesDecoded = 0;
  if (m_codec == CODEC_MP3) ret = MP3Decode(data, &m_bytesLeft, m_outBuff, 0);
  if (ret == 0) lastRet = 0;
  bytesDecoded = len - m_bytesLeft;
  // log_i("bytesDecoded %i", bytesDecoded);
  if (bytesDecoded == 0) { // unlikely framesize
    if (audio_info) audio_info("framesize is 0, start decoding again");
    m_f_playing = false; // seek for new syncword

    // we're here because there was a wrong sync word
    // so skip two sync bytes and seek for next
    return 2;
  }
  if (ret) { // Error, skip the frame...
    m_f_playing = false; // seek for new syncword
    if (count == 0) {
      printDecodeError(ret);
    }
    if (lastRet == ret) count++;
    if (count > 10) {
      count = 0;  // more than 10 errors
      lastRet = 0;
    }
    if (ret != lastRet) {
      lastRet = ret; count = 0;
    }
    return bytesDecoded;
  }
  else { // ret==0
    if (m_codec == CODEC_MP3) {
      if (MP3GetChannels() != lastChannels) {
        setChannels(MP3GetChannels());
        lastChannels = MP3GetChannels() ;
        sprintf(chbuf, "Channels=%i", MP3GetChannels() );
        if (audio_info) audio_info(chbuf);
      }
      if (MP3GetSampRate() != lastSampleRate) {
        setSampleRate(MP3GetSampRate());
        lastSampleRate = MP3GetSampRate();
        sprintf(chbuf, "SampleRate=%i", MP3GetSampRate());
        if (audio_info) audio_info(chbuf);
      }
      if ((int) MP3GetSampRate() == 0) { // error can't be
        if (audio_info) audio_info("SampleRate=0, try new frame");
        m_f_playing = false;
        return 1;
      }
      if (MP3GetBitsPerSample() != lastBitsPerSeconds) {
        lastBitsPerSeconds = MP3GetBitsPerSample();
        setBitsPerSample(MP3GetBitsPerSample());
        sprintf(chbuf, "BitsPerSample=%i", MP3GetBitsPerSample());
        if (audio_info) audio_info(chbuf);
      }
      if (MP3GetBitrate() != lastBitRate) {
        if (lastBitRate == 0) {
          sprintf(chbuf, "BitRate=%i", MP3GetBitrate());
          if (audio_info) audio_info(chbuf);
        }
        lastBitRate = MP3GetBitrate();
        m_bitRate = MP3GetBitrate();
      }
      m_validSamples = MP3GetOutputSamps() / lastChannels;
    }
  }
  compute_audioCurrentTime(bytesDecoded);

  // CAL
  m_curSample = 0;

  while (m_validSamples) {
    playChunk();
  }
  return bytesDecoded;
}

//---------------------------------------------------------------------------------------------------------------------
void AudioPlayer::compute_audioCurrentTime(int bd) {
  static uint16_t bitrate_counter = 0;
  static int      old_bitrate = 0;
  static uint32_t sum_bitrate = 0;
  static boolean  f_firstFrame = true;
  static boolean  f_CBR = true; // constant bitrate

  if (m_bitRate > 0) {
    if (m_avr_bitrate == 0) {
      bitrate_counter = 0;
      old_bitrate = 0;
      sum_bitrate = 0;
      f_firstFrame = true;
      f_CBR = true;
    }
    if (old_bitrate != m_bitRate) {
      if (!f_firstFrame && bitrate_counter == 0) {
        if (audio_info) audio_info("VBR recognized, audioFileDuration is estimated");
        f_CBR = false;
      }
    }
    old_bitrate = m_bitRate;
    if (bitrate_counter < 100 && !f_CBR) { // average of the first 100 values
      bitrate_counter++;
      sum_bitrate += m_bitRate;
      m_avr_bitrate = sum_bitrate /  bitrate_counter;
    }
    if (f_firstFrame) {
      m_avr_bitrate = m_bitRate; // if CBR set m_avr_bitrate only once
    }
    f_firstFrame = false;
    m_audioCurrentTime += (float) bd * 8 / m_bitRate;
  }
}
//---------------------------------------------------------------------------------------------------------------------
void AudioPlayer::printDecodeError(int r) {
  String e = "";
  if (m_codec == CODEC_MP3) {
    switch (r) {
      case ERR_MP3_NONE:                 e = "NONE";                  break;
      case ERR_MP3_INDATA_UNDERFLOW:     e = "INDATA_UNDERFLOW";      break;
      case ERR_MP3_MAINDATA_UNDERFLOW:   e = "MAINDATA_UNDERFLOW";    break;
      case ERR_MP3_FREE_BITRATE_SYNC:    e = "FREE_BITRATE_SYNC";     break;
      case ERR_MP3_OUT_OF_MEMORY:        e = "OUT_OF_MEMORY";         break;
      case ERR_MP3_NULL_POINTER:         e = "NULL_POINTER";          break;
      case ERR_MP3_INVALID_FRAMEHEADER:  e = "INVALID_FRAMEHEADER";   break;
      case ERR_MP3_INVALID_SIDEINFO:     e = "INVALID_SIDEINFO";      break;
      case ERR_MP3_INVALID_SCALEFACT:    e = "INVALID_SCALEFACT";     break;
      case ERR_MP3_INVALID_HUFFCODES:    e = "INVALID_HUFFCODES";     break;
      case ERR_MP3_INVALID_DEQUANTIZE:   e = "INVALID_DEQUANTIZE";    break;
      case ERR_MP3_INVALID_IMDCT:        e = "INVALID_IMDCT";         break;
      case ERR_MP3_INVALID_SUBBAND:      e = "INVALID_SUBBAND";       break;
      default: e = "ERR_UNKNOWN";
    }
    sprintf(chbuf, "MP3 decode error %d : %s", r, e.c_str());
    if (audio_info) audio_info(chbuf);
  }
}

//---------------------------------------------------------------------------------------------------------------------
uint32_t AudioPlayer::getFileSize() {
  if (!audiofile) return 0;
  return audiofile.size();
}
//---------------------------------------------------------------------------------------------------------------------
uint32_t AudioPlayer::getFilePos() {
  if (!audiofile) return 0;
  return audiofile.position();
}
//---------------------------------------------------------------------------------------------------------------------
uint32_t AudioPlayer::getAudioFileDuration() {
  if (m_f_localfile) {
    if (!audiofile) return 0;

    if ( 0 == m_audioFileDuration ) // calculate only once per file
    {
      uint32_t fileSize = getFileSize();
      uint32_t bitRate = m_avr_bitrate; //MP3GetBitrate();
      if (0 < bitRate)
      {
        m_audioFileDuration = 8 * (fileSize - m_id3Size) / bitRate;
      }
    }
    return m_audioFileDuration;
  }
  return 0;
}
//---------------------------------------------------------------------------------------------------------------------
uint32_t AudioPlayer::getAudioCurrentTime() { // return current time in seconds
  return (uint32_t) m_audioCurrentTime;
}
//---------------------------------------------------------------------------------------------------------------------
bool AudioPlayer::setFilePos(uint32_t pos) {
  if (!audiofile) return false;
  return audiofile.seek(pos);
}
//---------------------------------------------------------------------------------------------------------------------
bool AudioPlayer::audioFileSeek(const int8_t speed) {
  bool retVal = false;
  if (audiofile && speed) {
    retVal = true; //
    int32_t steps = 20 * speed;

    if ((steps < 0) && m_f_running) {
      // fast-rewind faster than fast-forward because we still playing
      steps = steps * 2;
    }
    uint32_t newPos = audiofile.position() + steps * MP3GetBitsPerSample();
    newPos = (newPos < m_id3Size) ? m_id3Size : newPos;
    newPos = (newPos >= audiofile.size()) ? audiofile.size() - 1 : newPos;
    if (audiofile.position() != newPos) {
      retVal = audiofile.seek(newPos);
    }
  }
  return retVal;
}


//---------------------------------------------------------------------------------------------------------------------
uint32_t AudioPlayer::inBufferFilled() {
  return InBuff.bufferFilled();
}

//---------------------------------------------------------------------------------------------------------------------
uint32_t AudioPlayer::inBufferFree() {
  return InBuff.freeSpace();
}
