/*
   AudioPlayer.h

    Created on: Oct 26,2018
    Updated on: Aug 22,2020
        Author: Wolle (schreibfaul1)

    Modifications and Implementation: Craig A. Lindley
    Last Update: 12/19/2024
*/

#ifndef AUDIOPLAYER_H_
#define AUDIOPLAYER_H_

#include "Arduino.h"
#include "SPI.h"
#include "SD.h"
#include "FS.h"

extern __attribute__((weak)) void audio_info(const char*);
extern __attribute__((weak)) void audio_id3data(const char*); //ID3 metadata
extern __attribute__((weak)) void audio_id3image(File&, const int); //ID3 metadata image
extern __attribute__((weak)) void audio_eof_mp3(const char*); //end of mp3 file
extern __attribute__((weak)) void audio_eof_stream(const char*); // The webstream comes to an end

//----------------------------------------------------------------------------------------------------------------------

class AudioBuffer {
    // AudioBuffer will be allocated in PSRAM, If PSRAM not available or has not enough space AudioBuffer will be
    // allocated in FlashRAM with reduced size
    //
    //                      m_readPtr                 m_writePtr                 m_endPtr
    //                           |<------dataLength------->|<------ writeSpace ----->|
    //                           ▼                         ▼                         ▼
    // ---------------------------------------------------------------------------------------------------------------
    // |                       <--m_buffSize-->                                      |      <--m_resBuffSize -->     |
    // ---------------------------------------------------------------------------------------------------------------
    // |<------freeSpace-------->|                         |<------freeSpace-------->|
    //
    //
    //
    // if the space between m_readPtr and buffend < 1600 bytes copy data from the beginning to resBuff
    // so that the mp3 frame is always completed
    //
    //                               m_writePtr                 m_readPtr        m_endPtr
    //                                    |<-------writeSpace-1 --->|<--dataLength-->|
    //                                    ▼                         ▼                ▼
    // ---------------------------------------------------------------------------------------------------------------
    // |                       <--m_buffSize-->                                      |      <--m_resBuffSize -->     |
    // ---------------------------------------------------------------------------------------------------------------
    // |<---  ------dataLength--  ------>|<-------freeSpace------->|
    //
    //

  public:
    AudioBuffer();                      // constructor
    ~AudioBuffer();                     // frees the buffer
    size_t   init();                    // set default values
    size_t   freeSpace();               // number of free bytes to overwrite
    size_t   writeSpace();              // space fom writepointer to bufferend
    size_t   bufferFilled();            // returns the number of filled bytes
    void     bytesWritten(size_t bw);   // update writepointer
    void     bytesWasRead(size_t br);   // update readpointer
    uint8_t* writePtr();                // returns the current writepointer
    uint8_t* readPtr();                 // returns the current readpointer
    uint32_t getWritePos();             // write position relative to the beginning
    uint32_t getReadPos();              // read position relative to the beginning
    void     resetBuffer();             // restore defaults

  protected:
    const size_t m_buffSizePSRAM = 300000; // most webstreams limit the advance to 100...300Kbytes
    const size_t m_buffSizeRAM   = 1600 * 5;
    size_t       m_buffSize      = 0;
    size_t       m_freeSpace     = 0;
    size_t       m_writeSpace    = 0;
    size_t       m_dataLength    = 0;
    size_t       m_resBuffSize   = 1600; // reserved buffspace, >= one mp3 frame
    uint8_t*     m_buffer        = NULL;
    uint8_t*     m_writePtr      = NULL;
    uint8_t*     m_readPtr       = NULL;
    uint8_t*     m_endPtr        = NULL;
    bool         m_f_start       = true;

};
//----------------------------------------------------------------------------------------------------------------------

class AudioPlayer : private AudioBuffer {

    AudioBuffer InBuff; // instance of input buffer

  public:
    AudioPlayer();
    ~AudioPlayer();
    bool connecttoFS(fs::FS &fs, String file);
    bool connecttoSD(String sdfile);
    void loop();
    uint32_t getFileSize();
    uint32_t getFilePos();

    uint32_t getSampleRate() {
      return m_sampleRate;
    }

    uint8_t getBitsPerSample() {
      return m_bitsPerSample;
    }

    uint8_t getChannels() {
      return m_channels;
    }

    /**
       @brief Get the audio file duration in seconds

       @return uint32_t file duration in seconds, 0 if no file active
    */
    uint32_t getAudioFileDuration();
    /**
       @brief Get the current plying time in seconds

       @return uint32_t current second of audio file, 0 if no file active
    */
    uint32_t getAudioCurrentTime();
    bool setFilePos(uint32_t pos);
    /**
       @brief audioFileSeek seeks the file in both directions

       @param[in] speed
            speed > 0 : fast-forward
            speed < 0 : fast-rewind
       @return true if audio file active and speed is valid, otherwise false
    */
    bool audioFileSeek(const int8_t speed);
    void stopSong();
    /**
       @brief pauseResume pauses current playback

       @return true if audio file or stream is active, false otherwise
    */
    bool pauseResume();

    void setVolume(uint8_t vol);
    uint8_t getVolume();

    bool isRunning() {
      return m_f_running;
    }
    uint32_t inBufferFilled(); // returns the number of stored bytes in the inputbuffer
    uint32_t inBufferFree();   // returns the number of free bytes in the inputbuffer

  private:
    void reset(); // free buffers and set defaults
    void processLocalFile();
    int  sendBytes(uint8_t *data, size_t len);
    void compute_audioCurrentTime(int bd);
    void printDecodeError(int r);
    void readID3Metadata();
    bool playChunk();
    int16_t Gain(int16_t s);
    void handlebyte(uint8_t b);

    //---------------------------------------------------------------------------------------------------------------------
    bool setSampleRate(uint32_t sampRate) {
      m_sampleRate = sampRate;
      return true;
    }

    //---------------------------------------------------------------------------------------------------------------------
    bool setBitsPerSample(int bits) {
      if ( (bits != 16) && (bits != 8) ) return false;
      m_bitsPerSample = bits;
      return true;
    }

    //---------------------------------------------------------------------------------------------------------------------
    bool setChannels(int ch) {
      if ((ch < 1) || (ch > 2)) return false;
      m_channels = ch;
      return true;
    }

  private:
    enum : int { APLL_AUTO = -1, APLL_ENABLE = 1, APLL_DISABLE = 0 };
    enum : int { EXTERNAL_I2S = 0, INTERNAL_DAC = 1, INTERNAL_PDM = 2 };
    enum : int { CODEC_NONE = 0, CODEC_WAV = 1, CODEC_MP3 = 2, CODEC_AAC = 4, CODEC_FLAC = 5};
    typedef enum { LEFTCHANNEL = 0, RIGHTCHANNEL = 1 } SampleIndex;




    File            audiofile;    // @suppress("Abstract class cannot be instantiated")
    char            chbuf[256];
    int             m_id3Size = 0;                  // length id3 tag
    int             m_bytesLeft = 0;
    uint32_t        m_bitRate = 0;                  // current bitrate given from decoder
    uint32_t        m_avr_bitrate;                  // average bitrate, median computed by VBR
    uint8_t         m_rev = 0;                      // revision
    String          m_audioName = "";               // the name of the file
    bool            m_f_unsync = false;
    bool            m_f_exthdr = false;             // ID3 extended header
    bool            m_f_localfile = false ;         // Play from local mp3-file
    bool            m_f_running = false;
    bool            m_f_stream = false;             // Set false if stream is lost
    uint8_t         m_codec = CODEC_NONE;           //
    //    bool            m_f_playing = false;            // valid mp3 stream recognized
    bool            m_f_psram = false;              // set if PSRAM is availabe
    uint32_t        m_audioFileDuration = 0;
    float           m_audioCurrentTime = 0;
    uint32_t        m_sampleRate    = 44100;
    uint8_t         m_bitsPerSample = 16;           // bitsPerSample
    uint8_t         m_channels      = 2;

};

#endif /* AUDIO_H_ */
